/* $Id: paintwindow.c,v 1.2 2006/01/28 16:46:10 cegger Exp $
***************************************************************************

   X server for LibGGI - Graphics acceleration

   Copyright (C) 1997	   Michael Krause	[rawstyle@ms.demo.org]
   Copyright (C) 1998-1999 Marcus Sundberg	[marcus@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************************************
*/

#include "screen.h"


/*
 * xggiPaintWindow - accelerated replacement for cfbPaintWindow
 *
 */
void xggiPaintWindow(WindowPtr pWin, RegionPtr prgn, int what)
{
	int scr;
	ScreenPtr pScreen = pWin->drawable.pScreen;
	int nBox = REGION_NUM_RECTS(prgn);
	BoxPtr pBox = REGION_RECTS(prgn);
	int fg = -1;
	PixmapPtr pPix = NULL;
	cfbPrivWin *pPrivWin;

	pPrivWin = cfbGetWindowPrivate(pWin);

	scr = xggiGetScreenIdx(pScreen);

	switch (what) {
	case PW_BACKGROUND:
		switch(pWin->backgroundState) {
		case None: return;
		case ParentRelative:
			do {
				pWin = pWin->parent;
			} while(pWin->backgroundState == ParentRelative);
			(*pWin->drawable.pScreen->PaintWindowBackground)(pWin, prgn, what);
			return;
		case BackgroundPixel:
			fg = pWin->background.pixel;
			break;
		case BackgroundPixmap:
			pPix = pWin->background.pixmap;
			break;
		}	/* switch */
		break;
	case PW_BORDER:
		if (pWin->borderIsPixel) {
			fg = pWin->border.pixel;
		} else {	/* pixmap */
			pPix = pWin->border.pixmap;
		}	/* if */
		break;
	default:
		return;
	}

#if 0	/* Too many small boxes are significantly slower than software rendering */
	if (!pPix) {
		ggi_pixel oldfg;

		DBUF_PROLOGUE(scr);
		ggiGetGCForeground(MASTER_VISUAL_VIS(scr), &oldfg);
		ggiSetGCForeground(MASTER_VISUAL_VIS(scr), (ggi_pixel)(fg));

		while (nBox--) {
			ggiDrawBox(MASTER_VISUAL_VIS(scr),
				pBox->x1, pBox->y1, pBox->x2 - pBox->x1,
				pBox->y2 - pBox->y1);
			pBox++;
		}	/* while */

		ggiSetGCForeground(MASTER_VISUAL_VIS(scr), oldfg);
		DBUF_EPILOGUE(scr);
		return;

	} else if (pPix->drawable.type == DRAWABLE_PIXMAP) {
#else
	if ((!pPix) || (pPix->drawable.type == DRAWABLE_PIXMAP)) {
#endif

		switch (SCREEN_ROOTDEPTH(scr)) {
		case 1:
			mfbPaintWindow(pWin, prgn, what);
			break;
		case 8:
			cfbPaintWindow(pWin, prgn, what);
			break;
		case 15:
		case 16:
			cfb16PaintWindow(pWin, prgn, what);
			break;
		case 24:
		case 32:
			if (MASTER_VISUAL_BPP(scr) == 24) {
				cfb24PaintWindow(pWin, prgn, what);
			} else {
				cfb32PaintWindow(pWin, prgn, what);
			}	/* if */
			break;
		}	/* switch */
		return;

	} else if (PIXMAP_HAS_MASTER_VISUAL_BPP(pPix, scr)) {

		int x, y, phaseY, phaseX, phaseXB, height, width, blit_w;
		int xorg, yorg;
		int pHeight = pPix->drawable.height;
		int pWidth = pPix->drawable.width;
		unsigned char *pSrc;
		int srcwidth = pPix->devKind;

		WindowPtr pBgWin = pWin;

		if (what == PW_BORDER) {
			for (pBgWin = pWin;
				pBgWin->backgroundState == ParentRelative;
				pBgWin = pBgWin->parent);
		}	/* if */

		xorg = pBgWin->drawable.x;
		yorg = pBgWin->drawable.y;

#ifdef PANORAMIX
		if (!noPanoramiXExtension) {
			int index = pScreen->myNum;
			if (WindowTable[index] == pBgWin) {
				xorg -= panoramiXdataPtr[index].x;
				yorg -= panoramiXdataPtr[index].y;
			}	/* if */
		}	/* if */
#endif

		DBUF_PROLOGUE(scr);
		while (nBox--) {
			x = pBox->x1;
			y = pBox->y1;
			phaseY = (pBox->y1 - yorg) % pHeight;
			if (phaseY < 0) phaseY += pHeight;
			phaseX = (x - xorg) % pWidth;
			pSrc = (unsigned char *)pPix->devPrivate.ptr +
				phaseX * (pPix->drawable.bitsPerPixel + 7) / 8;
			if (phaseX < 0) phaseX += pWidth;
			height = pBox->y2 - pBox->y1;
			width = pBox->x2 - x;

			while(1) {
				int ch = height, cp = phaseY, cy = y;
    
				blit_w = pWidth - phaseX;
				if (blit_w > width) blit_w = width;

				while (ch > 0) {
					int h = MIN(pHeight - cp, ch);
        
					ggiPutBox(MASTER_VISUAL_VIS(scr),
						x, cy, blit_w, h,
						(void *)(pSrc + cp * srcwidth));

					cy += h;
					ch -= h;
					cp = 0;
				}	/* while */
        
				width -= blit_w; 
				if (!width) break;
				x += blit_w;
				phaseX = (phaseX + blit_w) % pWidth;
				phaseXB = phaseX * pPix->drawable.bitsPerPixel / 8;
			}	/* while */
			pBox++;
		}	/* while */

		DBUF_EPILOGUE(scr);
		return;
	}	/* if */

#ifdef XGGI_DEBUG
	fprintf(stderr, "xggiPaintWindow: unhandled case found!\n");
#endif
	return;

}	/* xggiPaintWindow */

